#=======================================================================
#	$Id: Redland.pm,v 1.1 2005/11/26 19:57:24 pythontech Exp $
#	Storage of triples using Redland
#	Copyright (C) 2005  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#=======================================================================
package PythonTech::Triples::Redland;
use strict;
use PythonTech::Triples;
use RDF::Redland;

our @ISA = qw(PythonTech::Triples);

sub new {
    my $class = shift;
    my $storage = new RDF::Redland::Storage("hashes", "test",
					    "new='yes',hash-type='memory'")
	or die "Cannot create Redland hash memory storage\n";
    my $model = new RDF::Redland::Model($storage, "")
	or die "Cannot create Redland model\n";
    my $self = {
	storage => $storage,
	model => $model,
    };
    bless $self, $class;
    return $self;
}

sub configure {
}

#-----------------------------------------------------------------------
#	Assert a statement
#-----------------------------------------------------------------------
sub add {
    my($self, $subj,$pred,$obj) = @_;
    my $robj;
    if ($obj =~ /^[=^]/) {
	# Literal
	$robj = new RDF::Redland::Node(substr($obj,1));
    } else {
	$robj = $obj;
    }
    my $stmt = new RDF::Redland::Statement($subj,
					   $pred,
					   $robj);
    $self->{'model'}->add_statement($stmt);
}

# FIXME remove()

#-----------------------------------------------------------------------
#	Iterate over matching statements
#-----------------------------------------------------------------------
sub iterate {
    my($self, $subj,$pred,$obj,$cb) = @_;
    my @res;
    my $pat = join("",map {defined($_) || '?'} $subj,$pred,$obj);
    my $robj;
    if ($obj =~ /^=/) {
	# Literal
	$robj = new RDF::Redland::Node(substr($obj,1));
    } else {
	$robj = $obj;
    }
    my $stmt = new RDF::Redland::Statement($subj,
					   $pred,
					   $robj);
    my $stream = $self->{'model'}->find_statements($stmt);
    while ($stream && ! $stream->end) {
	my $match = $stream->current;
	# Create list of elements being queried for
	my @spo;
	push @spo, &_n2t($match->subject);
	push @spo, &_n2t($match->predicate);
	push @spo, &_n2t($match->object);
	my $res = $cb->(@spo);
	last if $res;
	$stream->next;
    }
    return;
}

sub _n2t {
    my($node) = @_;
    if ($node->is_resource) {
	return $node->uri->as_string;
    } elsif ($node->is_literal) {
	return "=" . $node->literal_value;
    } else {
	return "";		# Blank node
    }
}

#-----------------------------------------------------------------------
#	Read N-Triples file
#-----------------------------------------------------------------------
sub read_ntriples {
    my($self, $filename) = @_;
    my $parser = new RDF::Redland::Parser('ntriples', 'text/plain')
	or die "Failed to create Redland N-Triples parser\n";
    my $stream = $parser->parse_as_stream(new RDF::Redland::URI("file:$filename"), undef);
    my $model = $self->{'model'};
    while (! $stream->end) {
	my $stmt = $stream->current;
	$model->add_statement($stmt);
	$stream->next;
    }
}

#-----------------------------------------------------------------------
#	Read RDF/XML file
#-----------------------------------------------------------------------
sub read_rdfxml {
    my($self, $filename) = @_;
    my $parser = new RDF::Redland::Parser('raptor', 'application/rdf+xml')
	or die "Failed to create Redland RDF/XML parser\n";
    my $stream = $parser->parse_as_stream(new RDF::Redland::URI("file:$filename"), undef);
    my $model = $self->{'model'};
    while (! $stream->end) {
	my $stmt = $stream->next;
	$model->add_statement($stmt);
    }
}

#-----------------------------------------------------------------------
#	Read DBD files
#-----------------------------------------------------------------------
sub read_bdb {
    my($self, $filestub) = @_;
    my $dbstorage = new RDF::Redland::Storage("hashes", $filestub,
					      "new='no',hash-type='bdb'")
	or die "Cannot create Redland DBD storage\n";
    my $dbmodel = new RDF::Redland::Model($dbstorage, "")
	or die "Cannot create Redland model\n";
    # Copy statements into our model
    $self->{'model'}->add_statements($dbmodel->as_stream);
}

#-----------------------------------------------------------------------
#	Output model as RDF/XML to file (default stdout)
#-----------------------------------------------------------------------
sub write_rdfxml {
    my($self, $filename) = @_;
    my $serializer = new RDF::Redland::Serializer("rdfxml")
	or die "Cannot create Redland serializer\n";
    $serializer->serialize_model_to_file($filename || "/dev/stdout", undef, 
					 $self->{'model'});
}

#-----------------------------------------------------------------------
#	Output model as N-Triples to file (default stdout)
#-----------------------------------------------------------------------
sub write_ntriples {
    my($self, $filename) = @_;
    my $serializer = new RDF::Redland::Serializer("ntriples")
	or die "Cannot create Redland serializer\n";
    $serializer->serialize_model_to_file($filename || "/dev/stdout", undef, 
					 $self->{'model'});
}

#-----------------------------------------------------------------------
#	Output model to DB files
#-----------------------------------------------------------------------
sub write_bdb {
    my($self, $filestub) = @_;
    my $dbstorage = new RDF::Redland::Storage("hashes", $filestub,
					      "new='yes',hash-type='bdb'")
	or die "Cannot create Redland DBD storage\n";
    my $dbmodel = new RDF::Redland::Model($dbstorage, "")
	or die "Cannot create Redland model\n";
    # Copy statements into new model
    $dbmodel->add_statements($self->{'model'}->as_stream);
}

1;
